/*
 * CoverFinder - Find and download cover images from Amazon
 * Copyright (C) 2007 - 2008 - Sven Salzwedel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */
#include <gio/gio.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <errno.h>
#include <sys/stat.h>

#include "cf-search-result-amazon.h"
#include "cf-provider-amazon.h"
#include "cf-provider.h"
#include "cf-util.h"

#define AWS_BASEURL "http://ecs.amazonaws.%s/onca/xml?Service=AWSECommerceService"
#define AWS_OPERATION "&Operation=ItemSearch"
#define AWS_RESPONSES "&ResponseGroup=ItemAttributes,Images"
#define AWS_CONDITION "&Condition=All"
#define AWS_KEY "&AWSAccessKeyId=0VYQWKPNYF3M39Q9D502"
#define AWS_SEARCH_KEYWORDS "&Keywords=%s"
#define AWS_SEARCH_TITLE "&Title=%s"
#define AWS_SEARCH_ARTIST "&Artist=%s"
#define AWS_INDEX "&SearchIndex=Music"
#define AWS_PAGE "&ItemPage=%" G_GUINT64_FORMAT
#define AWS_URL AWS_BASEURL AWS_OPERATION AWS_RESPONSES AWS_CONDITION AWS_KEY AWS_INDEX

#define AWS_XML_ITEM_ATTRIB (xmlChar *) "ItemAttributes"
#define AWS_XML_ITEM_ATTRIB_TITLE (xmlChar *) "Title"
#define AWS_XML_ITEM_ATTRIB_ARTIST (xmlChar *) "Artist"
#define AWS_XML_ITEM_ATTRIB_BINDING (xmlChar *) "Binding"
#define AWS_XML_ITEM_SMALLIMG (xmlChar *) "SmallImage"
#define AWS_XML_ITEM_MEDIUMIMG (xmlChar *) "MediumImage"
#define AWS_XML_ITEM_LARGEIMG (xmlChar *) "LargeImage"
#define AWS_XML_ITEM_IMG_URL (xmlChar *) "URL"

#define AWS_XML_BINDING_AUDIO_CD (xmlChar *) "Audio CD"
#define AWS_XML_BINDING_CD_AUDIO (xmlChar *) "CD audio"
#define AWS_XML_BINDING_CD (xmlChar *) "CD"
#define AWS_XML_BINDING_DVD (xmlChar *) "DVD"
#define AWS_XML_BINDING_LP (xmlChar *) "LP Record"

#define READ_SIZE 4096

typedef struct
{
	gchar *url;
	gint width;
	gint height;
} CfSearchResultAmazonImage;

enum
{
	AMAZON_SMALL_IMAGE,
	AMAZON_MEDIUM_IMAGE,
	AMAZON_LARGE_IMAGE,
	AMAZON_NUM_IMAGE_TYPES
};

enum
{
	AMAZON_NOPROP, AMAZON_NAME, AMAZON_AWS_TLD
};

static GQuark error_domain;

static void cf_provider_amazon_query (GInputStream *istream,
                                      GOutputStream *ostream,
                                      GError **error)
{
	gsize read, written;
	guchar buf[READ_SIZE];
	GError *my_error = NULL;

	while ((read = g_input_stream_read (istream, buf, READ_SIZE, NULL,
	                                    &my_error)) != 0) {
		if (my_error != NULL) {
			g_propagate_error (error, my_error);
			return;
		}
		if (!g_output_stream_write_all (ostream, buf, read, &written, NULL,
		                                error)) {
			return;
		}
	}
}

static void cf_provider_amazon_search_result_add_image_uri (CfSearchResultAmazon *result,
                                                            xmlNode *children,
                                                            GError **error)
{
	xmlNode *node;
	xmlChar *val;
	gchar *uri = NULL;

	for (node = children; node != NULL; node = node->next) {
		val = xmlNodeGetContent (node);
		if (xmlStrEqual (node->name, AWS_XML_ITEM_IMG_URL)) {
			uri = g_strdup ((gchar *) val);
		}
		xmlFree (val);
	}

	g_object_set (G_OBJECT (result), "image-uri", uri, NULL);
	g_free (uri);
}

static void cf_provider_amazon_search_result_add_item_attributes (CfSearchResultAmazon *result,
                                                                  xmlNode *children)
{
	xmlNode *node;
	xmlChar *val;

	for (node = children; node != NULL; node = node->next) {
		val = xmlNodeGetContent (node);
		if (xmlStrEqual (node->name, AWS_XML_ITEM_ATTRIB_TITLE)) {
			g_object_set (G_OBJECT (result), "title", g_strdup ((gchar *) val), NULL);
		} else if (xmlStrEqual (node->name, AWS_XML_ITEM_ATTRIB_ARTIST)) {
			g_object_set (G_OBJECT(result), "artist", g_strdup ((gchar *) val), NULL);
		} else if (xmlStrEqual (node->name, AWS_XML_ITEM_ATTRIB_BINDING)) {
			gint type;

			if (xmlStrEqual (val, AWS_XML_BINDING_AUDIO_CD)
			    || xmlStrEqual (val, AWS_XML_BINDING_CD)
			    || xmlStrEqual (val, AWS_XML_BINDING_CD_AUDIO)) {
				type = AMAZON_TYPE_AUDIOCD;
			} else if (xmlStrEqual (val, AWS_XML_BINDING_DVD)) {
				type = AMAZON_TYPE_DVD;
			} else if (xmlStrEqual (val, AWS_XML_BINDING_LP)) {
				type = AMAZON_TYPE_LP;
			} else {
				//				g_print ("Didn't find a match for type: %s\n", val);
				type = AMAZON_TYPE_OTHER;
			}
			g_object_set (G_OBJECT (result), "type", type, NULL);
		}
		xmlFree (val);
	}
}

static void cf_provider_amazon_dowload_images (CfProviderAmazon *self,
                                               GError **error)
{
	GSList *lresult;
	CfSearchResultAmazon *result;
	gchar *uri, *cache_dir = NULL, *cache_root_dir, *basename, *dest_filename;
	gint num_results, i;
	gdouble fraction;
	GFile *source, *dest;

	g_return_if_fail (error != NULL && *error == NULL);

	cache_root_dir = cf_get_cache_dir ();
//	if (g_mkdir_with_parents (cache_dir, S_IRWXU) == -1) {
//		*error = g_error_new (error_domain, 0,
//		                      "Failed to create cache directory: %s\n",
//		                      g_strerror (errno));
//		goto ret;
//	}

	cf_provider_notify_progress (CF_PROVIDER (self), 0.0, "Downloading images from Amazon ...");
	num_results = g_slist_length (self->results);
	for (lresult = self->results, i = 1; lresult != NULL; lresult
	    = lresult->next, i++) {

		result = CF_SEARCH_RESULT_AMAZON (lresult->data);

		g_object_get (G_OBJECT (result), "image-uri", &uri, NULL);
		source = g_file_new_for_uri (uri);
		basename = g_file_get_basename (source);

		cache_dir = g_build_filename (cache_root_dir, basename, NULL);
		g_free (basename);
		if (g_mkdir_with_parents (cache_dir, S_IRWXU) == -1) {
			*error
			    = g_error_new (error_domain, 0,
			                   "Failed to create cache directory `%s': %s\n",
			                   cache_dir, g_strerror (errno));
			g_object_unref (source);
			goto ret;
		}

		dest_filename = g_build_filename (cache_dir, "cover.jpg", NULL);
		dest = g_file_new_for_path (dest_filename);

		g_free (uri);
		g_free (dest_filename);

		if (!g_file_query_exists (dest, NULL)) {
			g_file_copy (source, dest, G_FILE_COPY_NONE, NULL, NULL, NULL,
			             error);
		}

		g_object_unref (source);

		if (*error != NULL) {
			g_object_unref (dest);
			goto ret;
		}

		g_free (cache_dir);
		g_object_set (G_OBJECT (result), "image", dest, NULL);
		cf_provider_notify_result (CF_PROVIDER (self), CF_SEARCH_RESULT (result));
		fraction = (gdouble) i / (gdouble) num_results;
		cf_provider_notify_progress (CF_PROVIDER (self), fraction, "Downloading images from Amazon ...");
	}

	ret:

	g_free (cache_root_dir);
	cf_provider_notify_progress (CF_PROVIDER (self), 0.0, NULL);
}

static void cf_provider_amazon_parse_query_result (CfProviderAmazon *self,
                                                   gpointer data,
                                                   gsize size,
                                                   GError **error)
{
	xmlDoc *doc = NULL;
	xmlXPathContext *ctx = NULL;
	xmlXPathObject *obj = NULL;
	xmlNode *node = NULL;
	gint i, image_type;
	CfSearchResultAmazon *result;
	gchar *image_url;

	doc = xmlParseMemory (data, size);
	if (doc == NULL) {
		g_printerr ("Amazon Error: Invalid query result from Amazon.");
		goto ret;
	}
	ctx = xmlXPathNewContext (doc);
	if (xmlXPathRegisterNs (
	                        ctx,
	                        (xmlChar *) "a",
	                        (xmlChar *) "http://webservices.amazon.com/AWSECommerceService/2005-10-05")
	    == -1) {
		g_printerr (
		            "Amazon Error: Registration of namespace in query result failed.");
		goto ret;
	}
	obj
	    = xmlXPathEvalExpression (
	                              (xmlChar *) "/a:ItemSearchResponse/a:Items/a:Item",
	                              ctx);

	if (xmlXPathNodeSetIsEmpty (obj->nodesetval)) {
		goto ret;
	}

	for (i = 0; i < obj->nodesetval->nodeNr; i++) {
		node = obj->nodesetval->nodeTab[i];
		result = cf_search_result_amazon_new ();
		image_url = NULL;
		image_type = -1;
		for (node = node->children; node != NULL; node = node->next) {
			if (xmlStrEqual (node->name, AWS_XML_ITEM_ATTRIB)) {
				cf_provider_amazon_search_result_add_item_attributes (
				                                                      result,
				                                                      node->children);
			} else if (xmlStrEqual (node->name, AWS_XML_ITEM_LARGEIMG)) {
				cf_provider_amazon_search_result_add_image_uri (result,
				                                                node->children,
				                                                error);
				image_type = AMAZON_LARGE_IMAGE;
			} else if (xmlStrEqual (node->name, AWS_XML_ITEM_MEDIUMIMG)) {
				if (image_type <= AMAZON_MEDIUM_IMAGE) {
					cf_provider_amazon_search_result_add_image_uri (
					                                                result,
					                                                node->children,
					                                                error);
					image_type = AMAZON_MEDIUM_IMAGE;
				}
			} else if (xmlStrEqual (node->name, AWS_XML_ITEM_SMALLIMG)) {
				if (image_type <= AMAZON_SMALL_IMAGE) {
					cf_provider_amazon_search_result_add_image_uri (
					                                                result,
					                                                node->children,
					                                                error);
					image_type = AMAZON_SMALL_IMAGE;
				}
			}
			if (error == NULL || *error != NULL)
				return;
		}

		if (image_type == -1) {
			g_object_unref (result);
			continue;
		}
		self->results = g_slist_prepend (self->results, result);
	}

	self->results = g_slist_reverse (self->results);

	ret:

	cf_provider_notify_progress (CF_PROVIDER(self), 1.0, NULL);
	xmlFreeDoc (doc);
	xmlXPathFreeContext (ctx);
	xmlXPathFreeObject (obj);
}

static gchar *cf_provider_amazon_uri_escape (gchar *token)
{
	GError *error = NULL;
	gchar *converted, *escaped;
	gsize written;

	if ((converted = g_convert (token, -1, "ISO-8859-1", "UTF-8", NULL,
	                            &written, &error)) == NULL) {
		g_warning ("Failed to convert %s: %s\n", token, error->message);
		converted = g_strdup (token); // Hm, leave it ...
	}

	escaped = g_uri_escape_string (converted, NULL, FALSE);
	g_free (converted);
	return escaped;
}

static gchar *cf_provider_amazon_build_query_uri (CfProviderAmazon *self,
                                                  gint flavor_id,
                                                  CfSearch *search)
{
	gchar *escaped;
	GString *uri;

	uri = g_string_new (NULL);
	g_string_append_printf (uri, AWS_BASEURL, self->aws_tld);
	g_string_append (uri, AWS_OPERATION);
	g_string_append (uri, AWS_RESPONSES);
	g_string_append (uri, AWS_CONDITION);
	g_string_append (uri, AWS_INDEX);
	g_string_append (uri, AWS_KEY);

	if (search->keywords != NULL) {
		escaped = cf_provider_amazon_uri_escape (search->keywords);
		g_string_append_printf (uri, AWS_SEARCH_KEYWORDS, escaped);
		g_free (escaped);
	}
	if (search->title != NULL) {
		escaped = cf_provider_amazon_uri_escape (search->title);
		g_string_append_printf (uri, AWS_SEARCH_TITLE, escaped);
		g_free (escaped);
	}
	if (search->artist != NULL) {
		escaped = cf_provider_amazon_uri_escape (search->artist);
		g_string_append_printf (uri, AWS_SEARCH_ARTIST, escaped);
		g_free (escaped);
	}
	return g_string_free (uri, FALSE);
}

static void cf_provider_amazon_perform_search (CfProvider *iface,
                                               gint flavor_id,
                                               CfSearch *search)
{
	gchar *uri;
	GFile *file = NULL;
	GError *error = NULL;
	GFileInputStream *istream = NULL;
	GOutputStream *ostream = NULL;
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (iface);
	gpointer data;
	gsize size;

	g_return_if_fail (!(search->keywords == NULL && search->title == NULL && search->artist == NULL));
	g_return_if_fail (CF_IS_SEARCH (search));

	cf_provider_notify_progress (CF_PROVIDER(self), 0.0, "Querying Amazon Database");
	uri = cf_provider_amazon_build_query_uri (self, flavor_id, search);
	file = g_file_new_for_uri (uri);
	istream = g_file_read (file, NULL, &error);
	if (error != NULL) {
		g_printerr ("Error reading file: %s\n", error->message);
		goto ret;
	}

	ostream = g_memory_output_stream_new (NULL, 0, g_realloc, g_free);
	cf_provider_amazon_query (G_INPUT_STREAM (istream), ostream, &error);
	if (error != NULL) {
		g_printerr ("Error reading file: %s\n", error->message);
		goto ret;
	}

	data = g_memory_output_stream_get_data (G_MEMORY_OUTPUT_STREAM (ostream));
	size = g_memory_output_stream_get_size (G_MEMORY_OUTPUT_STREAM (ostream));
	cf_provider_amazon_parse_query_result (self, data, size, &error);
	if (error != NULL) {
		g_printerr ("Getting query results failed: %s\n", error->message);
		goto ret;
	}

	cf_provider_amazon_dowload_images (self, &error);
	if (error != NULL) {
		g_printerr ("Downloading images failed: %s\n", error->message);
	}

	ret:

	g_free (uri);
	if (error != NULL)
		g_error_free (error);
	if (file != NULL)
		g_object_unref (file);
	if (istream != NULL)
		g_object_unref (istream);
	if (ostream != NULL)
		g_object_unref (ostream);
}

/* interface implementations */

static void cf_provider_amazon_free_result (gpointer data, gpointer user_data)
{
	CfSearchResult *result = CF_SEARCH_RESULT (data);
	g_object_unref (G_OBJECT (result));
}

static void cf_provider_amazon_free_results (CfProvider *iface)
{
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (iface);

	if (self->results == NULL)
		return;

	g_slist_foreach (self->results, cf_provider_amazon_free_result, NULL);
	g_slist_free (self->results);
	self->results = NULL;
}

static GSList *cf_provider_amazon_get_results (CfProvider *iface)
{
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (iface);

	return self->results;
}

static gchar *cf_provider_amazon_get_name (CfProvider *iface)
{
	gchar *name;

	g_object_get (G_OBJECT (iface), "name", &name, NULL);
	return name;
}

static void cf_provider_amazon_interface_init (gpointer g_iface,
                                               gpointer iface_data)
{
	CfProviderInterface *iface = (CfProviderInterface *) g_iface;

	iface->get_name = cf_provider_amazon_get_name;
	iface->perform_search = cf_provider_amazon_perform_search;
	iface->free_results = cf_provider_amazon_free_results;
	iface->get_results = cf_provider_amazon_get_results;
}

static void cf_provider_amazon_get_property (GObject *object,
                                             guint property_id,
                                             GValue *value,
                                             GParamSpec *pspec)
{
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (object);

	switch (property_id) {
	case AMAZON_NAME:
		g_value_set_string (value, self->name);
		break;
	case AMAZON_AWS_TLD:
		g_value_set_string (value, self->aws_tld);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void cf_provider_amazon_set_property (GObject *object,
                                             guint property_id,
                                             const GValue *value,
                                             GParamSpec *pspec)
{
	switch (property_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

/* object/class initialization */

G_DEFINE_TYPE_EXTENDED (CfProviderAmazon, cf_provider_amazon,
	G_TYPE_OBJECT, 0, G_IMPLEMENT_INTERFACE (CF_TYPE_PROVIDER,
		cf_provider_amazon_interface_init))

static void cf_provider_amazon_finalize (GObject *object)
{
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (object);
	if (G_OBJECT_CLASS (cf_provider_amazon_parent_class)->finalize)
		G_OBJECT_CLASS (cf_provider_amazon_parent_class)->finalize (object);

	cf_provider_amazon_free_results (CF_PROVIDER (self));
	g_free (self->name);
	g_free (self->aws_tld);
}

static void cf_provider_amazon_class_init (CfProviderAmazonClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GParamSpec *spec;

	object_class->get_property = cf_provider_amazon_get_property;
	object_class->set_property = cf_provider_amazon_set_property;
	object_class->finalize = cf_provider_amazon_finalize;

	spec = g_param_spec_string ("name", NULL, NULL, NULL, G_PARAM_READABLE);
	g_object_class_install_property (object_class, AMAZON_NAME, spec);

	spec = g_param_spec_string ("aws-tld", NULL, NULL, NULL, G_PARAM_READABLE);
	g_object_class_install_property (object_class, AMAZON_AWS_TLD, spec);

	error_domain = g_quark_from_string ("AmazonProvider");
}

static void cf_provider_amazon_init (CfProviderAmazon *self)
{
	self->results = NULL;
	self->aws_tld = NULL;
	self->name = NULL;
}

CfProviderAmazon*
cf_provider_amazon_new (const gchar *tld, const gchar *name)
{
	CfProviderAmazon *self = CF_PROVIDER_AMAZON (g_object_new (CF_TYPE_PROVIDER_AMAZON, NULL));

	self->aws_tld = g_strdup (tld);
	self->name = g_strdup (name);
	return self;
}
